
/* 
 * A sample implementation of integer "measurable" lists
 * Implementation by delegation to integer lists
 */


object MList {

  def empty ():MList = new MListEmpty()

  def cons (v:Int,l:MList):MList = new MListCons(v,l)
}


abstract class MList extends List {

  def isEmpty ():Boolean
  def first ():Int
  def rest ():MList
  def find (f:Int):Boolean

  def append (M:MList):MList

  def length ():Int
  def sum ():Int
}



class MListEmpty () extends MList {

  // delegate for empty list

  val del:List = List.empty()


  // some methods can be delegated to the empty list

  def isEmpty ():Boolean = del.isEmpty()
  def first ():Int = del.first()
  def find (f:Int):Boolean = del.find(f)
  override def toString ():String = del.toString()


  // some methods require a downcast from List to MList

  def downcast (l:List):MList = l match {
    case l2:MList => l2
    case _ => throw new RuntimeException("Illegal downcast to MList")
  }

  def rest ():MList = downcast(del.rest())


  // other methods cannot actually be delegated (Exercise: why?)
  
  // DOESN'T WORK: def append (M:MList):MList = downcast(del.append(M))
  def append (M:MList):MList = M


  // bridge methods

  def append (M:List):List = del.append(M)


  // actual functionality added by MListEmpty

  val accLength:Int = 0
  val accSum:Int = 0

  def length ():Int = accLength
  def sum ():Int = accSum
}


class MListCons (v:Int,r:MList) extends MList {

  // delegate for cons list

  val del:List = List.cons(v,r)   // we *know* r is an MList
 

  // some methods can be delegated to the cons list

  def isEmpty ():Boolean = del.isEmpty()
  def first ():Int = del.first()
  def find (f:Int):Boolean = del.find(f)
  override def toString ():String = del.toString()


  // some methods require a downcast from List to MList

  def downcast (l:List):MList = l match {
    case l2:MList => l2
    case _ => throw new RuntimeException("Illegal downcast to MList")
  }

  def rest ():MList = downcast(del.rest())


  // other methods cannot actually be delegated (Exercise: why?)
  
  // DOESN'T WORK: def append (M:MList):MList = downcast(del.append(M))
  def append (M:MList):MList = MList.cons(v,r.append(M))


  // bridge methods

  def append (M:List):List = del.append(M)


  // actual functionality added by MListCons

  val accLength:Int = 1+r.length()
  val accSum:Int = v+r.sum()

  def length ():Int = accLength
  def sum ():Int = accSum
}
